#ifndef __DS485_SOCKET_SERVER_IMPL__
#define __DS485_SOCKET_SERVER_IMPL__

#include "base/ds_packet.h"

#if HAVE_VISIBILITY
#define NETLIB_EXPORTED __attribute__((visibility("default")))
#else
#define NETLIB_EXPORTED
#endif

/* only 4 bits for transaction id */
#define MAX_CLIENTS     (1 << 4)

/**
 * How long threads might sleep until checking
 * server->active again
 * will delay shutdown by this many seconds
 */
#define MAX_THREAD_SLEEP           1 /* sec */

/**
 * When to consider a receiver thread to be stuck
 */
#define CLIENT_INACTIVE_THRESHOLD   30 /* sec */

enum {
    CL_POLL = 0,
    CL_RECV = 1,
    CL_SERVER_RECEIVE = 2,
    CL_PROCESS = 3,
    CL_CACHE_SNOOP_START = 10,
    CL_CACHE_SNOOP_BUDDY_LOCK = 11,
    CL_QUICK_REPLY = 20,
    CL_SHUTDOWN = 100,
    CL_CLOSED = 101,
};

/**
 * Client session
 */
typedef struct _dS485dClient {
    struct _dS485dClient*   next;
    struct dS485_server     *server;

    int fd;                   /* client socket descriptor */
    int rxLen;
    uint8_t  *rxBuffer;

    int transactionId;        /* serves like dsuid of the client, same for all requests of this client */
    pthread_t tid;            /* thread id */

    uint8_t apirev;           /* client API revision */
    uint8_t rxfilter;         /* reception packet filter */

    /*
     * Only lock the socket itself, not the whole structure, tbd
     * http://lwn.net/Articles/180387/
     *
     * Locks also access to field ts_activity
     */
    pthread_mutex_t lock;
    char peer_name[25];

    time_t ts_detach;
    time_t ts_activity;
    int last_command;
    int m_state; /* CL_POLL, CL_SHUTDOWN, ... */
} dS485dClient;

/* CAUTION: requires cl->lock to be held */
void _ds485n_client_detach(dS485dClient *cl);

/**
 * Server
 */
struct dS485_server
{
    dS485dClient* clSessions;
    int transid_pool[MAX_CLIENTS]; /* transaction id pool */
    int client_count;
    int client_limit;     /* do not accept more than this */
    pthread_rwlock_t clientListRWLock;

    uint64_t rxPacketNum;
    uint64_t txPacketNum;
    bool active;
    bool shutdown_complete;
    bool upperPath;

    /* server socket file descriptor */
    int sfd;
    struct dS485_server_callbacks cb;

    pthread_t watchdog_tid;
    int wdctrl_pipe[2];  /* unidirectional 0:recv, 1:send */
    int ctrl_pipe[2];    /* unidirectional 0:recv, 1:send */
};


/* CAUTION: functions prefixed with '_' require holding a lock */

/**
 * TCP is stream oriented, packetize with tag-length-value
 * _ds485n_pu_write/_ds485n_tlv_write call this function internally
 */
int _ds485n_tlv_write(dS485dClient *cl, uint8_t cmd, const void *data, int dlen);

/**
 * requests from upper dsm-api layer, not really ds485-netlib function
 * could be called container write
 */
int _ds485n_dsmapi_write(dS485dClient *cl, uint8_t cmd,
                            const ds485_container_t *cont);
/**
 * routable data packets
 */
int _ds485n_pu_write(dS485dClient *cl, const packet_union_t *packet);

/**
 * Forward packet to server and all peers incl. ourselves if promiscous
 */
int ds485n_pu_receive(struct dS485_server *server, packet_union_t *packet,
                      dS485dClient *sender);

/**
 * parse tag-length-value chunk of data
 */
int ds485_netlib_packet_parser(dS485dClient *cl, uint8_t* data);

/**
 * Create v1 bus change message
 */
void ds485n_create_bus_change(ds485_container_t *cont, dsuid_t *dsuid,
                              int flags);

/**
 * Create v1 bus state message
 */
void ds485n_create_link_state(ds485_container_t *_cont, bool _link);

#endif
